# -*- coding: utf-8 -*-

"""
SOAP request and response.
"""

import json
import dataclasses
from pathlib import Path
import xml.etree.ElementTree as ET

import requests

from .exc import SOAPResponseParseError


# ------------------------------------------------------------------------------
# Soap Request and Response
# ------------------------------------------------------------------------------
path_xml = Path(__file__).absolute().parent / "execute-command.xml"

# default soap request headers
_SOAP_REQUEST_HEADERS = {"Content-Type": "application/xml"}
_SOAP_REQUEST_XML_TEMPLATE = path_xml.read_text(encoding="utf-8")
DEFAULT_USERNAME = "admin"
DEFAULT_PASSWORD = "admin"
DEFAULT_HOST = "localhost"
DEFAULT_PORT = 7878


@dataclasses.dataclass
class Base:
    """
    Base class for :class:`SOAPRequest` and :class:`SOAPResponse`.
    """

    @classmethod
    def from_dict(cls, dct: dict):
        """
        Construct an object from a dict.
        """
        return cls(**dct)

    def to_dict(self) -> dict:
        """
        Convert the object to a dict.
        """
        return {k: v for k, v in dataclasses.asdict(self).items() if v is not None}

    @classmethod
    def from_json(cls, json_str: str):
        """
        Construct an object from a JSON string.
        """
        return cls.from_dict(json.loads(json_str))

    def to_json(self) -> str:  # pragma: no cover
        """
        Convert the object to a JSON string.
        """
        return json.dumps(self.to_dict())


@dataclasses.dataclass
class SOAPRequest(Base):
    """
    :class:`~acore_soap_app.agent.impl.SOAPRequest` is a dataclass to represent
    the SOAP XML request.

    Usage example

    .. code-block:: python

        # this code only works in where the worldserver is running
        >>> request = SOAPRequest(command=".server info")
        >>> response = request.send()
        >>> response.to_json()
        {
            "body": "<?xml version=\"1.0\" encoding=\"UTF-8\"?><SOAP-ENV:Envelope xmlns:SOAP-ENV=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:SOAP-ENC=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:xsi=\"http://www.w3.org/1999/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/1999/XMLSchema\" xmlns:ns1=\"urn:AC\"><SOAP-ENV:Body><ns1:executeCommandResponse><result>AzerothCore rev. 85311fa55983 2023-03-25 22:36:05 +0000 (master branch) (Unix, RelWithDebInfo, Static)&#xD;Connected players: 0. Characters in world: 0.&#xD;Connection peak: 0.&#xD;Server uptime: 54 minute(s) 3 second(s)&#xD;Update time diff: 10ms, average: 10ms.&#xD;</result></ns1:executeCommandResponse></SOAP-ENV:Body></SOAP-ENV:Envelope>",
            "message": "AzerothCore rev. 85311fa55983 2023-03-25 22:36:05 +0000 (master branch) (Unix, RelWithDebInfo, Static)Connected players: 0. Characters in world: 0.Connection peak: 0.Server uptime: 54 minute(s) 3 second(s)Update time diff: 10ms, average: 10ms.",
            "succeeded": true
        }

    :param command: the command to execute.
    :param username: the in game GM account username, default "admin".
    :param password: the in game GM account password, default "admin".
    :param host: wow world server host, default "localhost".
    :param port: wow world server SOAP port, default 7878.

    More methods from base class:

    - :meth:`~Base.from_dict`
    - :meth:`~Base.to_dict`
    - :meth:`~Base.from_json`
    - :meth:`~Base.to_json`
    """

    command: str = dataclasses.field()
    username: str = dataclasses.field(default=DEFAULT_USERNAME)
    password: str = dataclasses.field(default=DEFAULT_PASSWORD)
    host: str = dataclasses.field(default=DEFAULT_HOST)
    port: int = dataclasses.field(default=DEFAULT_PORT)

    @property
    def endpoint(self) -> str:
        """
        Construct the Soap service endpoint URL.
        """
        return f"http://{self.username}:{self.password}@{self.host}:{self.port}/"

    def send(self) -> "SOAPResponse":  # pragma: no cover
        """
        Run soap command via HTTP request. This function "has to" be run on the
        game server and talk to the localhost. You should NEVER open SOAP port
        to public!
        """
        http_response = requests.post(
            self.endpoint,
            headers=_SOAP_REQUEST_HEADERS,
            data=_SOAP_REQUEST_XML_TEMPLATE.format(command=self.command),
        )
        return SOAPResponse.parse(http_response.text)


@dataclasses.dataclass
class SOAPResponse(Base):
    """
    :class:`~acore_soap_app.agent.impl.SOAPResponse` is a dataclass to represent
    the SOAP XML response.

    Usage:

    .. code-block:: python

        >>> res = SOAPResponse.parse(
        ...  '''
        ...      <?xml version="1.0" encoding="UTF-8"?><SOAP-ENV:Envelope
        ...      ...<result>Account created: test&#xD;</result>...</SOAP-ENV:Envelope>
        ...  '''
        ... )
        >>> res.message
        Account created: test
        >>> res.succeeded
        True

    :param body: the raw SOAP XML response
    :param message: if succeeded, it is the ``<result>...</result>`` part.
        if failed, it is the ``<faultstring>...</faultstring>`` part
    :param succeeded: a boolean flag to indicate whether the command is succeeded

    More methods from base class:

    - :meth:`~Base.from_dict`
    - :meth:`~Base.to_dict`
    - :meth:`~Base.from_json`
    - :meth:`~Base.to_json`
    """

    body: str = dataclasses.field()
    message: str = dataclasses.field()
    succeeded: bool = dataclasses.field()

    @classmethod
    def parse(cls, body: str) -> "SOAPResponse":
        """
        Parse the SOAP XML response.
        """
        root = ET.fromstring(body)

        results = list(root.iter("result"))
        if len(results):
            result = results[0]
            if result.text:
                message = result.text.strip()
            else:
                message = "No result"
            return cls(
                body=body.strip(),
                message=message,
                succeeded=True,
            )

        faultstrings = list(root.iter("faultstring"))
        if len(faultstrings):
            faultstring = faultstrings[0]
            if faultstring.text:
                message = faultstring.text.strip()
            else:
                message = "No fault string"
            return cls(
                body=body.strip(),
                message=message,
                succeeded=False,
            )

        # todo: add logic to handle SOAPCommandFailedError situation
        raise SOAPResponseParseError(f"Cannot parse the response: {body!r}")

    def print(self):  # pragma: no cover
        """
        Print the dataclass, ignore the raw response body.
        """
        print({"succeeded": self.succeeded, "message": self.message})



# -*- coding: utf-8 -*-

"""
Common utilities.
"""

import dataclasses

from ..request import (
    SOAPRequest,
    SOAPResponse,
    DEFAULT_USERNAME,
    DEFAULT_PASSWORD,
    DEFAULT_HOST,
    DEFAULT_PORT,
)


@dataclasses.dataclass
class GMCommandRequest:
    """
    todo: docstring
    """

    def to_command(self) -> str:
        """
        Build the GM command.

        See all GM commands at https://www.azerothcore.org/wiki/gm-commands
        """
        raise NotImplementedError

    def send(
        self,
        username: str = DEFAULT_USERNAME,
        password: str = DEFAULT_PASSWORD,
        host: str = DEFAULT_HOST,
        port: int = DEFAULT_PORT,
    ) -> SOAPResponse:
        req = SOAPRequest(
            command=self.to_command(),
            username=username,
            password=password,
            host=host,
            port=port,
        )
        return req.send()


@dataclasses.dataclass
class GMCommandResponse:
    """
    todo: docstring
    """

    @classmethod
    def from_soap_response(self, res: SOAPResponse):
        """
        Parse from soap response.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

"""
todo: docstring
"""

import re
import dataclasses

from ..request import SOAPResponse
from ..exc import SOAPResponseParseError

from .base import GMCommandRequest, GMCommandResponse


@dataclasses.dataclass
class SetAccountPasswordResponse(GMCommandResponse):
    """
    Parse the response message of ``.account set password ...`` command.
    """

    @classmethod
    def from_soap_response(cls, res: SOAPResponse):
        raise NotImplementedError


@dataclasses.dataclass
class SetAccountPasswordRequest(GMCommandRequest):
    """
    todo: docstring
    """

    account: str = dataclasses.field()
    new_password: str = dataclasses.field()

    def to_command(self) -> str:
        return f".account set password {self.account} {self.new_password} {self.new_password}"



# -*- coding: utf-8 -*-

"""
Usage:

    >>> import acore_soap.api as acore_soap
    >>> request = acore_soap.gm.CreateAccountRequest(
    ...     account="testacc",
    ...     password="testpass",
    ... )
    >>> soap_response = request.send()
    >>> response = acore_soap.gm.CreateAccountResponse.from_soap_response(soap_response)
"""

import re
import dataclasses

from ..request import SOAPResponse
from ..exc import SOAPResponseParseError

from .base import GMCommandRequest, GMCommandResponse


@dataclasses.dataclass
class CreateAccountResponse(GMCommandResponse):
    """
    Parse the response message of ``.account create ...`` command.
    """

    @classmethod
    def from_soap_response(cls, res: SOAPResponse):
        raise NotImplementedError


@dataclasses.dataclass
class CreateAccountRequest(GMCommandRequest):
    """
    todo: docstring
    """

    account: str = dataclasses.field()
    password: str = dataclasses.field()

    def to_command(self) -> str:
        return f".account create {self.account} {self.password}"



# -*- coding: utf-8 -*-

from acore_soap.gm.server_info import (
    SOAPResponse,
    ServerInfoResponse,
)


class TestServerInfoResponse:
    def test_parse_server_uptime(self):
        # fmt: off
        assert ServerInfoResponse.extract_server_uptime("Server uptime: 34 second(s)") == 34
        assert ServerInfoResponse.extract_server_uptime("Server uptime: 6 minute(s) 29 second(s)") == 389
        assert ServerInfoResponse.extract_server_uptime("Server uptime: 1 hour(s) 6 minute(s) 29 second(s)") == 3989
        assert ServerInfoResponse.extract_server_uptime("Server uptime: 1 day(s) 1 hour(s) 6 minute(s) 29 second(s)") == 90389
        # fmt: one

    def test_from_soap_response(self):
        soap_response = SOAPResponse(
            body="",
            message="AzerothCore rev. 278ee2a72836 2024-06-13 21:52:22 +0200 (master branch) (Unix, RelWithDebInfo, Static)\r\nConnected players: 1000. Characters in world: 700.\r\nConnection peak: 0.\r\nServer uptime: 34 second(s)\r\nUpdate time diff: 1ms. Last 500 diffs summary:\r\n- Mean: 1ms\r\n- Median: 1ms\r\n- Percentiles (95, 99, max): 2ms, 12ms, 17ms",
            succeeded=True,
        )
        res = ServerInfoResponse.from_soap_response(soap_response)
        assert res.connected_players == 1000
        assert res.characters_in_world == 700
        assert res.server_uptime == 34


if __name__ == "__main__":
    from acore_soap.tests import run_cov_test

    run_cov_test(__file__, "acore_soap.gm.server_info", preview=False)



# -*- coding: utf-8 -*-

from acore_soap import api


def test():
    _ = api
    _ = api.SOAPResponseParseError
    _ = api.SOAPCommandFailedError
    _ = api.SOAPRequest
    _ = api.SOAPResponse
    _ = api.gm.GMCommandRequest
    _ = api.gm.GMCommandResponse
    _ = api.gm.ServerInfoRequest
    _ = api.gm.ServerInfoResponse
    _ = api.gm.CreateAccountRequest
    _ = api.gm.CreateAccountResponse
    _ = api.gm.ChangeAccountPasswordRequest
    _ = api.gm.ChangeAccountPasswordResponse
    _ = api.gm.SetAccountPasswordRequest
    _ = api.gm.SetAccountPasswordResponse
    _ = api.gm.SetAccountGmlevelRequest
    _ = api.gm.SetAccountGmlevelResponse
    _ = api.gm.AddItemRequest
    _ = api.gm.AddItemResponse


if __name__ == "__main__":
    from acore_soap.tests import run_cov_test

    run_cov_test(__file__, "acore_soap.api", preview=False)



# -*- coding: utf-8 -*-

"""
todo: docstring
"""

import re
import dataclasses

from ..request import SOAPResponse
from ..exc import SOAPResponseParseError

from .base import GMCommandRequest, GMCommandResponse


@dataclasses.dataclass
class SetAccountGmlevelResponse(GMCommandResponse):
    """
    Parse the response message of ``.account set gmlevel ...`` command.
    """

    @classmethod
    def from_soap_response(cls, res: SOAPResponse):
        raise NotImplementedError


@dataclasses.dataclass
class SetAccountGmlevelRequest(GMCommandRequest):
    """
    :param account:
    :param gmlevel:
        - 0: SEC_PLAYER
        - 1: SEC_MODERATOR
        - 2: SEC_GAMEMASTER
        - 3: SEC_ADMINISTRATOR
    :param realm_id: -1 for all realms
    """

    account: str = dataclasses.field()
    gmlevel: int = dataclasses.field()
    realm_id: int = dataclasses.field()

    def to_command(self) -> str:
        return f".account set gmlevel {self.account} {self.gmlevel} {self.realm_id}"



# -*- coding: utf-8 -*-

"""
todo: docstring
"""

import re
import dataclasses

from ..request import SOAPResponse
from ..exc import SOAPResponseParseError

from .base import GMCommandRequest, GMCommandResponse


@dataclasses.dataclass
class AddItemResponse(GMCommandResponse):
    """
    Parse the response message of ``.additem ...`` command.
    """

    @classmethod
    def from_soap_response(cls, res: SOAPResponse):
        raise NotImplementedError


@dataclasses.dataclass
class AddItemRequest(GMCommandRequest):
    """
    todo: docstring
    """

    player: str = dataclasses.field()
    item_id: int = dataclasses.field()
    count: int = dataclasses.field(default=1)

    def to_command(self) -> str:
        return f".additem {self.player} {self.item_id} {self.count}"



# -*- coding: utf-8 -*-


class SOAPResponseParseError(ValueError):
    """
    raises when failed to parse the soap response.
    """


class SOAPCommandFailedError(ValueError):
    """
    raises when SOAP command failed
    """



# -*- coding: utf-8 -*-

"""
Usage::

    import acore_soap.gm.api as acore_soap_gm
"""

from .base import GMCommandRequest
from .base import GMCommandResponse
from .server_info import ServerInfoRequest
from .server_info import ServerInfoResponse
from .create_account import CreateAccountRequest
from .create_account import CreateAccountResponse
from .change_account_password import ChangeAccountPasswordRequest
from .change_account_password import ChangeAccountPasswordResponse
from .set_account_password import SetAccountPasswordRequest
from .set_account_password import SetAccountPasswordResponse
from .set_account_gmlevel import SetAccountGmlevelRequest
from .set_account_gmlevel import SetAccountGmlevelResponse
from .add_item import AddItemRequest
from .add_item import AddItemResponse



# -*- coding: utf-8 -*-

"""
Usage:

    >>> import acore_soap.api as acore_soap
    >>> request = acore_soap.gm.ServerInfoRequest()
    >>> soap_response = request.send()
    >>> response = acore_soap.gm.ServerInfoResponse.from_soap_response(soap_response)
    >>> response.connected_players
    1
    >>> response.characters_in_world
    1
    >>> response.server_uptime
    3600
"""

import re
import dataclasses

from ..request import SOAPResponse
from ..exc import SOAPResponseParseError

from .base import GMCommandRequest, GMCommandResponse


@dataclasses.dataclass
class ServerInfoResponse(GMCommandResponse):
    """
    Parse the response message of ``.server info`` command, extract

    :param connected_players: how many players are logged in.
    :param characters_in_world: how many player
    :param server_uptime: how long the server has been running (in seconds)
    """

    connected_players: int = dataclasses.field()
    characters_in_world: int = dataclasses.field()
    server_uptime: int = dataclasses.field()

    @staticmethod
    def extract_connected_players(message: str) -> int:
        """
        Extract the number of connected players from the response message.
        """
        res = re.findall(r"Connected players: (\d+)", message)
        if len(res) == 1:
            return int(res[0])
        else:  # pragma: no cover
            raise SOAPResponseParseError(message)

    @staticmethod
    def extract_characters_in_world(message: str) -> int:
        """
        Extract the number of characters in the world from the response message.
        """
        res = re.findall(r"Characters in world: (\d+)", message)
        if len(res) == 1:
            return int(res[0])
        else:  # pragma: no cover
            raise SOAPResponseParseError(message)

    @staticmethod
    def extract_server_uptime(message: str) -> int:
        """
        Extract the server uptime from the response message.
        """
        P_SECOND = re.compile(r"(\d+\s+second\(s\))")
        P_MINUTE = re.compile(r"(\d+\s+minute\(s\))")
        P_HOUR = re.compile(r"(\d+\s+hour\(s\))")
        P_DAY = re.compile(r"(\d+\s+day\(s\))")

        def extract(p: re.Pattern, message: str) -> int:
            res = re.findall(p, message)
            if len(res) == 1:
                return int(res[0].split()[0])
            elif len(res) == 0:
                return 0
            else:  # pragma: no cover
                raise ValueError()

        seconds = extract(P_SECOND, message)
        minutes = extract(P_MINUTE, message)
        hours = extract(P_HOUR, message)
        days = extract(P_DAY, message)
        total_seconds = seconds + minutes * 60 + hours * 3600 + days * 86400
        return total_seconds

    @classmethod
    def from_soap_response(cls, res: SOAPResponse):
        connected_players = cls.extract_connected_players(res.message)
        characters_in_world = cls.extract_characters_in_world(res.message)
        server_uptime = cls.extract_server_uptime(res.message)
        return cls(
            connected_players=connected_players,
            characters_in_world=characters_in_world,
            server_uptime=server_uptime,
        )


@dataclasses.dataclass
class ServerInfoRequest(GMCommandRequest):
    """
    Generate the command to get server information.
    """

    def to_command(self) -> str:
        return ".server info"



# -*- coding: utf-8 -*-

"""
Azerothcore World of Warcraft Soap request and response Python library.
"""

from ._version import __version__

__short_description__ = "Azerothcore World of Warcraft Soap request and response Python library."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__maintainer__ = "Sanhe Hu"
__maintainer_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"
